/*****************************************************************************
 *
 * grail - Gesture Recognition And Instantiation Library
 *
 * Copyright (C) 2010 Canonical Ltd.
 * Copyright (C) 2010 Henrik Rydberg <rydberg@bitmath.org>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include "grail-inserter.h"
#include "grail-recognizer.h"
#include "grail-impl.h"
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <malloc.h>
#include <errno.h>

static void tp_event(struct touch_dev *dev,
		     const struct input_event *ev)
{
	struct grail *ge = dev->priv;
	struct grail_impl *x = ge->impl;
	if (ev->type == EV_ABS) {
		if (ev->code == ABS_X)
			x->pointer_x = ev->value;
		if (ev->code == ABS_Y)
			x->pointer_y = ev->value;
	}
	evbuf_put(&x->evbuf, ev);
}

static int extra_filtered(const struct input_event *ev)
{
	if (ev->type == EV_ABS) {
		return 1;
	}
	if (ev->type == EV_KEY) {
		switch (ev->code) {
		case BTN_TOUCH:
		case BTN_TOOL_FINGER:
		case BTN_TOOL_DOUBLETAP:
		case BTN_TOOL_TRIPLETAP:
		case BTN_TOOL_QUADTAP:
		return 1;
		}
	}
	return 0;
}

static int handle_abs_events(struct grail *ge,
			      const struct input_event *syn,
			      int pointer)
{
	struct grail_impl *impl = ge->impl;
	struct input_event ev = *syn;
	int nevent = 0;
	if (pointer != impl->pointer_status) {
		ev.type = EV_KEY;
		ev.code = BTN_TOUCH;
		ev.value = pointer;
		ge->event(ge, &ev);
		impl->pointer_status = pointer;
		nevent++;
	}
	if (pointer) {
		ev.type = EV_ABS;
		ev.code = ABS_X;
		ev.value = impl->pointer_x;
		ge->event(ge, &ev);
		nevent++;
		ev.type = EV_ABS;
		ev.code = ABS_Y;
		ev.value = impl->pointer_y;
		ge->event(ge, &ev);
		nevent++;
	}
	return nevent;
}

static void tp_sync(struct touch_dev *dev,
		    const struct input_event *syn)
{
	struct input_event ev;
	struct grail *ge = dev->priv;
	struct grail_impl *impl = ge->impl;
	struct gesture_recognizer *gru = ge->gru;
	struct touch_frame *frame = &dev->frame;
	grail_mask_t filtered[DIM_EV_TYPE_BYTES];
	int pointer, nevent = 0;
	gin_frame_begin(ge, frame);
	gru_recognize(ge, frame);
	gin_frame_end(ge, filtered, sizeof(filtered), frame);
	pointer = gru->move.single && !grail_mask_get(filtered, EV_ABS);

	if (!ge->event) {
		evbuf_clear(&impl->evbuf);
		return;
	}
	if (!impl->filter_abs)
		nevent += handle_abs_events(ge, syn, pointer);
	while (!evbuf_empty(&impl->evbuf)) {
		evbuf_get(&impl->evbuf, &ev);
		if (extra_filtered(&ev))
			continue;
		ge->event(ge, &ev);
		nevent++;
	}
	if (nevent)
		ge->event(ge, syn);
}

void grail_filter_abs_events(struct grail *ge, int usage)
{
	struct grail_impl *x = ge->impl;
	x->filter_abs = usage;
}

int grail_open(struct grail *ge, int fd)
{
	struct grail_impl *x;
	int ret;
	x = calloc(1, sizeof(*x));
	if (!x)
		return -ENOMEM;
	ge->impl = x;

	ret = touch_dev_open(&x->dev, fd);
	if (ret)
		goto freemem;
	x->dev.event = tp_event;
	x->dev.sync = tp_sync;
	x->dev.priv = ge;

	ret = gin_init(ge);
	if (ret)
		goto freedev;

	ret = gru_init(ge);
	if (ret)
		goto freegin;

	return 0;
 freegin:
	gin_destroy(ge);
 freedev:
	touch_dev_close(&x->dev, fd);
 freemem:
	free(x);
	ge->impl = 0;
	return ret;
}

void grail_close(struct grail *ge, int fd)
{
	struct grail_impl *x = ge->impl;
	void *status;
	gru_destroy(ge);
	gin_destroy(ge);
	touch_dev_close(&x->dev, fd);
	free(ge->impl);
	ge->impl = 0;
}

int grail_idle(struct grail *ge, int fd, int ms)
{
	struct grail_impl *x = ge->impl;
	return touch_dev_idle(&x->dev, fd, ms);
}

int grail_pull(struct grail *ge, int fd)
{
	struct grail_impl *x = ge->impl;
	return touch_dev_pull(&x->dev, fd);
}

void grail_get_units(const struct grail *ge,
		     struct grail_coord *min, struct grail_coord *max)
{
	const struct touch_caps *caps = &ge->impl->dev.caps;
	min->x = caps->min_x;
	min->y = caps->min_y;
	max->x = caps->max_x;
	max->y = caps->max_y;
}

